/*
 * Copyright (C) 2002 Edscott Wilson Garcia
 * EMail: edscott@imp.mx
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */



#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif

#include <errno.h>
#ifdef HAVE_LIMITS_H
#include <limits.h>
#endif
#ifdef HAVE_MEMORY_H
#include <memory.h>
#endif
#ifdef HAVE_STDARG_H
#include <stdarg.h>
#elif HAVE_VARARGS_H
#include <varargs.h>
#endif
#include <stdio.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>
#include <glib.h>

#ifdef XFCE_SUPPORT
#include <libxfce4util/util.h>
#endif

#include "glade_gui.h"
#include "glade_support.h"
#include "glade_callbacks.h"

#include "constants.h"
#include "types.h"

#include "run.h"
#include "reg.h"
#include "input.h"
#include "misc.h"
#include "entry.h"


static GList *run_list = NULL;
static GList *reg_list = NULL;

#ifndef G_THREADS_IMPL_NONE
static void *hilo_func(void *d)
{
    char *returnSTR = NULL;
    char **argv = (char **)d;
    /*printf("DBG:Thread here\n");
       {int i;for (i=0;argv[i]!=NULL;i++)printf("arg[%d]=%s\n",i,argv[i]);} */

    {
	int status, child;
	child = fork();
	if(child < 0)
	    g_assert_not_reached();

	if(!child)
	{
	    if(!fork())
	    {
		execvp(argv[0], argv);
		_exit(1);
	    }
	    _exit(1);
	}
	else
	    wait(&status);
    }

    return(returnSTR);
}
#endif

int runv(GtkTreeView * treeview, char *argv[])
{
#ifndef G_THREADS_IMPL_NONE
    gpointer thread_return = NULL;
    GThread *thread;
#endif
    gchar *fullpath;

    fullpath=g_find_program_in_path(argv[0]);


    if(!fullpath || access(fullpath,X_OK)!=0)
    {
	/*printf("DBG:strlen(%s)=%d\n",fullpath,strlen(fullpath)); */
	/*if (strlen(fullpath))printf("DBG:access=%d\n",
	   access(fullpath,X_OK)); */
	print_status(treeview, "xf_ERROR_ICON", strerror(EACCES), " ", 
			argv[0], NULL);
	return FALSE;
    } else g_free(fullpath);
    
    chdir(get_selected_chdir(treeview));

#ifndef G_THREADS_IMPL_NONE
    /* initialize the thread system */
    if (!g_thread_supported())
        g_thread_init(NULL);

    if ((thread = g_thread_create((GThreadFunc)hilo_func, (gpointer)argv,
                    TRUE, NULL)) != NULL) {
        thread_return = g_thread_join(thread);
    }
    else
#endif
    {
	    int status, child;
	
        child = fork();
	if(child < 0)
	    g_assert_not_reached();
	if(!child)
	{
	    usleep(5000);
	    if(!fork())
	    {
	        usleep(5000);
		execvp(argv[0], argv);
		_exit(1);
	    }
    	    chdir("/");
	    _exit(1);
	}
	else {
    	    chdir("/");
	    wait(&status);
	}
    }
    return TRUE;

}

/* run without files associated (input)
   run with file associated: double click(input if !registered)
   openwith (menu): requires selection, can be multiple.
   */



int on_run(GtkTreeView * treeview, char *in_cmd, tree_entry_t * en, gboolean in_terminal, gboolean remember)
{
    char *p,*term= "xterm";
    GList *ee = run_list;
    int i;
    char *argv[MAX_ARGS], *cmd,*cmd_line;
    char app[_POSIX_PATH_MAX], arg[_POSIX_PATH_MAX];
    gboolean more_args,is_argv0=TRUE;

    cmd = g_strdup(in_cmd);
    cmd_line = g_strdup(in_cmd);

    i = 0;
    arg[0] = 0;
    if(in_terminal)
    {
	if (getenv("TERM") && strlen( getenv("TERM"))){
		gchar *c=g_find_program_in_path(getenv("TERM"));
		if (c && access(c,X_OK)==0) term = getenv("TERM");
		g_free(c);	
	}
		
	if(!term || strlen(term) == 0)
	    term = "xterm";
	argv[i++] = term;
	if(strcmp(term, "xterm") == 0)
	{
	    argv[i++] = "-T";
	    argv[i++] = in_cmd;
	    argv[i++] = "+mesg";
	    argv[i++] = "-aw";
	    argv[i++] = "+cm";
	    argv[i++] = "+dc";
	    if (getenv("XFFM_HOLD_XTERM") && strlen( getenv("XFFM_HOLD_XTERM"))) argv[i++] = "-hold";
	    argv[i++] = "-j";
	    argv[i++] = "-rightbar";
	    argv[i++] = "-s";
	    argv[i++] = "-sb";
	    argv[i++] = "-u8";
	    argv[i++] = "-wf";
	    argv[i++] = "-sl";
	    argv[i++] = "1024";
	}
	argv[i++] = "-e";
	sprintf(app, "%s", term);	/* in terminal */
	sprintf(arg, "%s", "-e");	/* in terminal */
    }


    /* remove leading and trailing whitespace */
    p=g_strstrip(cmd);

    if (*p) more_args=TRUE; else more_args=FALSE;
    while (more_args){
       char tok;
       if (*p=='\"'){  p++;  tok='\"';} else  tok=' ';
       for (argv[i]=p;*p && *p!=tok;p++);
       if (*p) {*p=0;p++;more_args=TRUE;} else more_args=FALSE;
#ifdef DEBUG
	    printf("DBG: argv[%d]=%s\n",i,argv[i]);
#endif
       if (!strlen(argv[i])) break;
       if (is_argv0){
	    is_argv0=FALSE;
 	    if(!in_terminal) sprintf(app,(tok==' ')?"%s":"\"%s\"", argv[i]);
	    else {
	       strcat(arg, " ");
	       if (tok!=' ') strcat(arg, "\"");
	       strcat(arg, argv[i]);
	       if (tok!=' ') strcat(arg, "\"");
	    }
	} else {
	       strcat(arg, " ");
	       if (tok!=' ') strcat(arg, "\"");
	       strcat(arg, argv[i]);
	       if (tok!=' ') strcat(arg, "\"");
	}
        if (i == MAX_ARGS-1){
	     argv[i]=NULL;
	     break;
        } 
        i++;
    }
    argv[i]=NULL;
			    
	    
#if 0
    if(strstr(cmd, " "))
    {
	argv[i] = strtok(cmd, " ");
	if(!in_terminal)
	    sprintf(app, "%s", argv[i]);
	else
	{
	    strcat(arg, " ");
	    strcat(arg, argv[i]);
	}
	i++;
	do
	{
	    argv[i] = strtok(NULL, " ");
	    if(!argv[i])
		break;
	    else
	    {
		strcat(arg, " ");
		strcat(arg, argv[i]);
	    }
	    i++;
	    if(i >= MAX_ARGS)
	    {
		argv[MAX_ARGS - 1] = NULL;
		break;
	    }

	}
	while(1);
    }
    else
    {
	argv[i] = cmd;
	if(!in_terminal)
	    sprintf(app, "%s", argv[i]);
	else
	{
	    strcat(arg, " ");
	    strcat(arg, argv[i]);
	}
	i++;
	argv[i] = NULL;
    }
#endif

    /* command argv done, only need to add the file path */
    if(en)
    {
	if(i < MAX_ARGS - 1)
	{
	    argv[i++] = en->path;
	    argv[i] = NULL;
	}
	/*{int j; for(j=0;j<i;j++) printf("DBG: argv[%d]=%s\n",j,argv[j]);} */
    }
    if(!runv(treeview, argv))
    {
	if(en)
	    g_free(cmd);
	return FALSE;
    }

    if(en) g_free(cmd);
    else {
        static void save_run_list(void);
	/*while(ee) {
	    if(strcmp(cmd, (char *)ee->data) == 0)
		break;
	    ee = ee->next;
	}
	if(ee) g_free(cmd)
	else */
	run_list = g_list_prepend(run_list, cmd_line);
	save_run_list();
    }
    if(en && remember)
    {
	char *sfx;
	if(!en->path)
	    g_assert_not_reached();
	sfx = strchr(en->path, '.');
	if (sfx && strchr(sfx,'/')) sfx=NULL;

	if(!sfx && strlen(en->path) > 1)
	{
	    sfx = strrchr(en->path, '/');
	    if(sfx) sfx++;
	}
	if(sfx)
	{
	    /*printf("DBG: adding %s %s\n",app,arg); */
	    reg_add_suffix(sfx, app, arg);
	    reg_save();
	}
    }

    return TRUE;
}


#if 0
static void save_run_list(void)
{				/* save last 13 entries */
    FILE *f;
    char fname[_POSIX_PATH_MAX];
    int i = 0;
    GList *g = run_list;
    xfce_get_userfile_r(fname, _POSIX_PATH_MAX-1, "xffm%cxffm.runlist",
		    G_DIR_SEPARATOR);
    f = fopen(fname, "w");
    if(!f)
	return;
    while(g)
    {
	fprintf(f, "%s\n", (char *)g->data);
	g = g->next;
	if(++i >= 13)
	    break;
    }
    fclose(f);
}
static GList *load_run_list(void)
{
    FILE *f;
    char fname[_POSIX_PATH_MAX];
    GList *g = NULL;
    xfce_get_userfile_r(fname, _POSIX_PATH_MAX-1, "xffm%cxffm.runlist",
		    G_DIR_SEPARATOR);
    f = fopen(fname, "r");
    if(!f)
	return NULL;
    while(fgets(fname, 255, f) && !feof(f))
    {
	fname[255] = 0;
	if(strstr(fname, "\n"));
	strtok(fname, "\n");
	g = g_list_append(g, g_strdup(fname));
    }
    fclose(f);
    return g;
}
#endif

static void save_run_list(void)
{				
    char fname[_POSIX_PATH_MAX];
    char *p;
    if (!run_list || !run_list->data || !strlen((char *)run_list->data))
	    return;
    p=(char *)run_list->data;
    xfce_get_userfile_r(fname, _POSIX_PATH_MAX-1, "xffm%cxffm.runlist.dbh",
		    G_DIR_SEPARATOR); 
    save_to_history(fname,p);
}
static GList *load_run_list(void)
{
    char fname[_POSIX_PATH_MAX];
    /* eliminate pre 4.0-beta3 runlist, if any */
    xfce_get_userfile_r(fname, _POSIX_PATH_MAX-1, "xffm%cxffm.runlist",
		    G_DIR_SEPARATOR);
    unlink(fname);
    xfce_get_userfile_r(fname, _POSIX_PATH_MAX-1, "xffm%cxffm.runlist.dbh",
		    G_DIR_SEPARATOR);
    return get_history_list(fname);
}

void set_run_combo(GtkTreeView * treeview)
{
    tree_details_t *tree_details = get_tree_details(treeview);
    GtkWidget *w;
    w = lookup_widget(tree_details->window, "input_combo");
    run_list = load_run_list();
    if(run_list)
	gtk_combo_set_popdown_strings(GTK_COMBO(w), run_list);
}

void set_reg_combo(GtkTreeView * treeview)
{
    char *p;
    tree_details_t *tree_details = get_tree_details(treeview);
    GtkWidget *w;
    GtkTreeIter iter;
    tree_entry_t *en;
    en = get_selected_entry(treeview, &iter);

    w = lookup_widget(tree_details->window, "input_combo");
    reg_list = reg_app_list();

    p = g_strdup(reg_app_by_file(en->path));
    reg_list = g_list_prepend(reg_list, p);

    if(reg_list)
    {
	gtk_combo_set_popdown_strings(GTK_COMBO(w), reg_list);
	reg_list = reg_app_list_free(reg_list);
    }

}

void double_click_run(tree_details_t *tree_details,
		tree_entry_t *en){
    int i = 0;
    char *argv[20];
    char *cmd;
    FILE *pipe;
    char *a=NULL;
    gboolean interm=FALSE;
    /* interm specified? */
    if (IS_IN_TERM(en->subtype)) interm=TRUE;
    else if (!IS_APP_TYPE(en->type)){
    /* is it a script? (not applicable to application stuff*/
     cmd = (char *)malloc(strlen("file \"xx\"") + strlen(en->path) + 1);
     if(!cmd) g_assert_not_reached();
     sprintf(cmd, "file \"%s\"", en->path);
     pipe = popen(cmd, "r");
     if(pipe)
     {
	char *p, line[1024];
	fgets(line, 1023, pipe);
	line[1023] = 0;
	pclose(pipe);
	if((p = strstr(line, ": ")) != NULL)
	{
	    p += 2;
	    if(strstr(p, "script")) interm=TRUE;
	}
     }
     g_free(cmd);
    }
    if (interm){
	argv[i] = "xterm";
	if (getenv("TERM") && strlen( getenv("TERM"))){
	    gchar *c=g_find_program_in_path(getenv("TERM"));
	    if (c && access(c,X_OK)==0) argv[i] = getenv("TERM");
	}	    

	if(strcmp(argv[i++], "xterm") == 0)
	{
	    argv[i++] = "-T";
	    argv[i++] = strrchr(en->path, '/') + 1;
	    argv[i++] = "+mesg";
	    argv[i++] = "-aw";
	    argv[i++] = "+cm";
	    argv[i++] = "+dc";
	    if (getenv("XFFM_HOLD_XTERM") && strlen( getenv("XFFM_HOLD_XTERM"))) argv[i++] = "-hold";
	    argv[i++] = "-j";
	    argv[i++] = "-rightbar";
	    argv[i++] = "-s";
	    argv[i++] = "-sb";
	    argv[i++] = "-u8";
	    argv[i++] = "-wf";
	    argv[i++] = "-sl";
	    argv[i++] = "1024";
	}
	argv[i++] = "-e";
    }
    argv[i++] = en->path;
    if (en->filter) { /* filter element is double used for arguments here */
	    char *b;
	    a=g_strdup(en->filter);
	    /*printf("DBG: arguments=%s\n",en->filter);*/
	    b=strtok(a," ");
	    if (!b) argv[i++]=a;
	    else while (b) {
		    argv[i++]=b;
		    b=strtok(NULL," ");
	    } 	    
    }
    argv[i] = NULL;
    print_status(tree_details->treeview, "xf_INFO_ICON",
		   _("Executing"), " : ", 
		    FILENAME(en), 
		   NULL);
    runv(tree_details->treeview, argv);
    g_free(a);
    return;
}

void double_click_open_with(tree_details_t *tree_details,
		tree_entry_t *en){
    /* open with */
    char *name, *sfx = NULL;
    char arguments[256];
    reg_t *prg;
    name = strrchr(en->path, '/');
    if(!name)
	g_assert_not_reached();
    if(strlen(name) > 1)
	name++;
    sfx = strchr(name, '.');
    if(!sfx)
	sfx = name;
    /*printf("DBG:looking for %s\n",sfx); */
    prg = reg_prog_by_suffix(sfx);
    if(prg)
    {
	sprintf(arguments, "%s %s", prg->app, (prg->arg) ? prg->arg : " ");
	if(on_run(tree_details->treeview, arguments, en, FALSE, FALSE))
	{
	    print_status(tree_details->treeview, "xf_INFO_ICON", 
			    arguments, 
			    NULL);
	}
	else
	{
	    print_status(tree_details->treeview, "xf_ERROR_ICON", 
			    strerror(EINVAL), arguments, 
			    NULL);
	}
	/*printf("DBG:%s %s\n",arguments,en->path); */
    }
    else
    {		/*if !registered, run directly */
	/*printf("DBG:regnot found: %s\n",en->path); */
	show_input(tree_details->treeview, RUN_DOUBLE_CLICK);
	print_status(tree_details->treeview, "xf_QUESTION_ICON", 
			_("Input requested"),NULL);
	/*if(IS_BOOKMARK_TYPE(en->type) && IS_DIR(en->type))
	{
	    GtkEntry *entry;
	    entry = (GtkEntry *) 
		    lookup_widget(tree_details->window, "input_entry");
	    gtk_entry_set_text(entry, "xffm");
		
	}*/
	    
    }
    return;
}



/* callbacks */
void tb_run(GtkButton * button, gpointer user_data)
{
    GtkTreeView *treeview = get_treeview((GtkWidget *) button);

/*   GtkWidget *w;
   w=lookup_widget(widget,"input_combo");
   if (run_list) gtk_combo_set_popdown_strings (GTK_COMBO (w),run_list);*/
    show_input(treeview, RUN_INPUT);
}

void on_open_with_activate(GtkMenuItem * menuitem, gpointer user_data)
{
    GtkTreeView *treeview = get_treeview((GtkWidget *) menuitem);
    tree_entry_t *en;
    GtkTreeIter iter;
    en = get_selected_entry(treeview, &iter);
    if(!en)
	show_input(treeview, RUN_INPUT);
    else
    {
	show_input(treeview, RUN_DOUBLE_CLICK);
	print_status(treeview, "xf_QUESTION_ICON", _("Input requested"), NULL);
    }


}

void on_run_activate(GtkMenuItem * menuitem, gpointer user_data)
{
    GtkTreeView *treeview = get_treeview((GtkWidget *) menuitem);
    show_input(treeview, RUN_INPUT);

}
